"""
:class:`DistributeFilesDataset`

https://github.com/rwth-i6/returnn/issues/1519
"""

from __future__ import annotations

from collections import deque
from typing import Union, Optional, Any, Callable, Sequence, Tuple, List, Dict
import os
import sys
import numpy
from returnn.log import log
from returnn.util import better_exchook
from returnn.util.basic import override_env_var, try_run
from returnn.util.literal_py_to_pickle import literal_eval
from returnn.util.multi_proc_non_daemonic_spawn import NonDaemonicSpawnContext
from returnn.config import SubProcCopyGlobalConfigPreInitFunc
from .basic import init_dataset, extend_dataset_dict_from_parent_dataset, DatasetSeq, RANDOM_SEED_OFFSET_ENV_VAR
from .cached2 import CachedDataset2

# noinspection PyProtectedMember
from multiprocessing.connection import Connection as mpConnection

_mp = NonDaemonicSpawnContext(process_pre_init_func=SubProcCopyGlobalConfigPreInitFunc())


__all__ = ["DistributeFilesDataset"]

Filename = str
FileTree = Union[Filename, Tuple["FileTree", ...], Dict[Any, "FileTree"], List["FileTree"]]


class DistributeFilesDataset(CachedDataset2):
    """
    Dataset that distributes files over subepochs and then creates a
    sub dataset for every sub epoch for a given (random) subset of the files.
    The sub dataset is user-defined via a function ``get_sub_epoch_dataset``.
    Thus, this dataset wraps the sub datasets.

    It is conceptually very similar to :class:`ConcatDataset` in the sense
    that it concatenates all the sub datasets together to form one larger dataset.

    This scheme allows to shuffle over the files,
    which makes shuffling much more efficient over a large dataset
    at the cost of no longer shuffling over the full dataset in every subepoch.
    Instead, the quality of the shuffle depends on the number of files the dataset is
    split into -- the more files per subepoch, the better.

    Additionally, this scheme allows to prefetch and cache the upcoming needed files,
    e.g. copying them from a NFS to local disk,
    when the local disk can not store the whole dataset,
    and/or when the local disk would allow for faster access.
    For that, the user can use :class:`returnn.util.file_cache.CachedFile`
    inside the returned dict from ``get_sub_epoch_dataset``.

    It was also designed with multi-GPU training in mind,
    where each GPU should get a set of files to work on,
    and the sub epochs across the GPUs would not be exactly of the same length.
    Whenever one GPU finishes its sub epoch, all remaining data on the other GPUs is dropped.
    First, our scheme tries to make the length of the sub epochs as equally distributed as possible
    after random shuffling by using the size of the files as proxy for the length of the contents.
    Further, due to the random shuffling, there should not be any bias in the data distribution.
    Specifically, we don't want that some data might be visited more often than others
    (at least its expected value should be the same).

    In case the dataset grows so large it is unreasonable to expect one worker to
    ever see all the data, this dataset can also shard the file list on a per-worker
    basis before distributing across subepochs.
    This behavior can be configured by setting the property ``"distrib_shard_files": True``.
    The dataset attempts to split the files as evenly as possible based on the file size.

    Example usage::

        def get_sub_epoch_dataset(files_subepoch: List[str]) -> Dict[str, Any]:
          from returnn.util.file_cache import CachedFile
          return {
            "class": "HDFDataset",
            "files": [CachedFile(fn) for fn in files_subepoch],
          }

        train = {
          "class": "DistributeFilesDataset",
          "files": [
            "/nfs/big_data_1.hdf",
            ...
          ],  # M files
          "get_sub_epoch_dataset": get_sub_epoch_dataset,
          "partition_epoch": P,  # P << M
        }

    Instead of a plain list of strings, you can also provide a list of any nested structures to keep
    multimodal data together::

        def get_sub_epoch_dataset(files_subepoch: List[Tuple[str, str]]) -> Dict[str, Any]:
          from returnn.util.file_cache import CachedFile

          alignments, features = tuple(zip(*files_subepoch)) # transpose

          return {
            "class": "MetaDataset",
            "data_map": {"classes": ("alignments", "data"), "data": ("features", "data")},
            "datasets": {
              "alignments": {
                "class": "HDFDataset",
                "files": alignments,
                "seq_ordering": "random",
              },
              "features": {
                "class": "HDFDataset",
                "files": features,
              },
            },
            "seq_order_control_dataset": "alignments",
          }

        train = {
          "class": "DistributeFilesDataset",
          "files": [
            ("/nfs/alignment_1.hdf", "/nfs/features_1.hdf"),
            ...
          ],  # M entries
          "get_sub_epoch_dataset": get_sub_epoch_dataset,
          "partition_epoch": P,  # P << M
        }

    In this case the file sizes for sub epoch distribution are summed up per list entry
    by iterating over the structure leaves.

    For some discussion, see https://github.com/rwth-i6/returnn/issues/1519 and
    https://github.com/rwth-i6/returnn/issues/1524.
    """

    def __init__(
        self,
        *,
        files: Union[List[FileTree], os.PathLike],
        get_sub_epoch_dataset: Callable[[List[FileTree]], Dict[str, Any]],
        preload_next_n_sub_epochs: int = 1,
        buffer_size: int = 1,
        distrib_shard_files: bool = False,
        _meta_info_cache: Optional[Dict[str, Any]] = None,
        _distrib_info: Optional[Dict[str, int]] = None,
        **kwargs,
    ):
        """
        :param files: the files to shuffle over, can also be a list of arbitrarily nested python objects
            to keep associated heterogeneous data together.
            When the list grows too large to be serialized into a RETURNN config, the list of files
            can also be specified as a path to a .txt file containing one file per line,
            or a python file containing the repr of a list of arbitrarily nested python objects,
            or a JSON file containing a list of arbitarily nested (JSON) objects.
        :param get_sub_epoch_dataset: callable which returns a dataset dict for a given subset of files
        :param preload_next_n_sub_epochs: how many sub epoch datasets to preload
        :param buffer_size: buffer size for each worker, amount of seqs to prefetch
        :param distrib_shard_files: set to true to shard the data across worker processes in
            distributed training scenaria
        :param _meta_info_cache: for internal use
        :param _distrib_info: for internal use
        """
        super().__init__(**kwargs)
        self.files = files
        self.get_sub_epoch_dataset = get_sub_epoch_dataset
        assert preload_next_n_sub_epochs >= 0
        self.preload_next_n_sub_epochs = preload_next_n_sub_epochs
        self.buffer_size = buffer_size
        self._file_sizes: Optional[Dict[str, int]] = None  # key -> size. for equal distribution across sub epochs
        self._data_keys: Optional[List[str]] = None
        self._num_seqs: Optional[int] = None

        self._files: Optional[List[FileTree]] = None  # files to use for this dataset
        self._workers: Dict[int, _WorkerProcParent] = {}  # epoch -> worker
        self._files_order_cache: Dict[int, List[List[FileTree]]] = {}  # full epoch (0-indexed) -> files order

        self.distrib_shard_files = distrib_shard_files
        if distrib_shard_files:
            if _distrib_info:
                # We're in a child process.
                # `_get_rank_and_size()` no longer works,
                # so we pass the info about the shards via a pickled property.
                # See also Dataset.__reduce__.
                # _num_shards and _shard_index are already set, so just check.
                assert (
                    self._shard_index == _distrib_info["_shard_index"]
                    and self._num_shards == _distrib_info["_num_shards"]
                )
            else:
                assert self._num_shards == 1 and self._shard_index == 0, (  # ensure defaults are set
                    f"{self}: Cannot use both dataset-sharding via properties _num_shards and _shard index "
                    f"and {self.__class__.__name__}'s own sharding implementation based on the trainings rank and size."
                )
                self._shard_index, self._num_shards = _get_rank_and_size()
        assert 0 <= self._shard_index < self._num_shards

        if _meta_info_cache:
            # This allows to skip the lazy init in self.initialize().
            # This can be used when pickling/deepcopying an instance of this dataset,
            # see Dataset.__reduce__.
            self.num_inputs = _meta_info_cache["num_inputs"]
            self.num_outputs = _meta_info_cache["num_outputs"]
            self.labels = _meta_info_cache["labels"]
            self._data_keys = _meta_info_cache["data_keys"]
            self._file_sizes = _meta_info_cache["file_sizes"]
            self._files = _meta_info_cache["files"]

    def initialize(self):
        """init"""
        self._lazy_init_num_outputs()
        super().initialize()

    def supports_sharding(self) -> bool:
        """:return: whether the dataset supports sharding based on the seq_order"""
        return True

    @property
    def _distrib_info(self):
        return {"_num_shards": self._num_shards, "_shard_index": self._shard_index}

    @property
    def _meta_info_cache(self):
        if not self.num_outputs:
            return None
        return {
            "num_inputs": self.num_inputs,
            "num_outputs": self.num_outputs,
            "labels": self.labels,
            "data_keys": self._data_keys,
            "file_sizes": self._file_sizes,
            "files": self._files,
        }

    def _uses_custom_distributed_sharding(self) -> bool:
        return self._num_shards > 1

    def _lazy_init_file_list(self):
        """
        The list of data files can either be provided as python list, or, if that grows
        too large, as path to a file containing the list.

        This function initializes the list of files from whatever was given as input.
        """
        if self._files is not None:
            return
        if isinstance(self.files, list):
            self._files = self.files
        elif isinstance(self.files, (str, os.PathLike)):
            _, ext = os.path.splitext(self.files)
            assert ext, f"{self}: no file extension on file list file {self.files}"
            if ext == ".txt":
                with open(self.files, "rt") as f:
                    stripped_lines = (line.strip() for line in f.readlines())
                    self._files = [line for line in stripped_lines if line and not line.startswith("#")]
            elif ext == ".json":
                import json

                with open(self.files, "rt") as f:
                    self._files = json.load(f)
            elif ext == ".py":
                with open(self.files, "rb") as f:
                    self._files = literal_eval(f.read())
            else:
                raise ValueError(f"{self}: type {ext} not supported as file list file")
            assert isinstance(self._files, list)
        else:
            raise ValueError(f"{self}: unsupported file list ({type(self.files)}: {self.files})")
        if len(self._files) < self.partition_epoch:
            raise ValueError(f"{self}: len(files) {len(self._files)} < partition_epoch {self.partition_epoch}")

    def _lazy_init_num_outputs(self):
        if self.num_outputs:
            return
        self._lazy_init_file_list()
        # First, we need to know the num_inputs, num_outputs, total_num_seqs, labels.
        # Init the dataset with the first file.
        dataset_dict = self._get_sub_dataset_dict(files=[self._files[0]])
        dataset = init_dataset(dataset_dict, extra_kwargs={"seq_ordering": "default"}, parent_dataset=self)
        self.num_inputs = dataset.num_inputs
        self.num_outputs = dataset.num_outputs
        self.labels = dataset.labels
        self._data_keys = dataset.get_data_keys()

    def _lazy_init_file_sizes(self):
        import tree

        if self._file_sizes:
            return
        self._lazy_init_file_list()
        self._file_sizes = {
            _get_key_for_file_tree(t): sum((os.path.getsize(fn) for fn in tree.flatten(t)), 0) for t in self._files
        }

    def __del__(self):
        if hasattr(self, "_workers"):  # might not be set after an early exception in __init__
            for k, worker in self._workers.items():
                try_run(worker.exit, kwargs={"join": False})
            self._workers.clear()

    def init_seq_order(self, epoch: Optional[int] = None, seq_list=None, seq_order=None) -> bool:
        """
        :param epoch:
        :param seq_list:
        :param seq_order:
        :return: whether the order changed (True is always safe to return)
        """
        super().init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        if seq_list is not None:
            raise Exception(f"{self}: seq_list not supported")
        if seq_order is not None:
            raise Exception(f"{self}: seq_order not supported")
        if epoch is None:
            self._num_seqs = 0
            return True

        self._lazy_init_file_list()
        self._lazy_init_file_sizes()

        full_epoch_0idx = (epoch - 1) // self.partition_epoch

        # Cleanup and fill _files_order_cache, also shard files across GPU workers
        for k in list(self._files_order_cache.keys()):
            if k < full_epoch_0idx:
                del self._files_order_cache[k]
        for ep_ in range(epoch, epoch + self.preload_next_n_sub_epochs + 1):
            full_epoch_0idx_ = (ep_ - 1) // self.partition_epoch
            if full_epoch_0idx_ in self._files_order_cache:
                continue
            if self.seq_ordering == "default":
                files_order_flat = self._files
            elif self.seq_ordering == "random":
                # when sharding, _get_random_seed_for_epoch makes sure to use a fixed
                # random_seed_offset
                rnd_seed = self._get_random_seed_for_epoch(full_epoch_0idx_ * self.partition_epoch + 1)
                random_generator = numpy.random.RandomState(rnd_seed)
                files_order_flat = list(self._files)
                random_generator.shuffle(files_order_flat)
            else:
                raise ValueError(f"{self}: seq_ordering {self.seq_ordering!r} not supported")
            file_bins = self._distribute_evenly_by_size(
                num_bins=self._num_shards * self.partition_epoch,
                file_sizes=self._file_sizes,
                files_order=files_order_flat,
            )
            self_index_base = self.partition_epoch * self._shard_index
            self_index_end = self_index_base + self.partition_epoch
            self._files_order_cache[full_epoch_0idx_] = file_bins[self_index_base:self_index_end]

        # Cleanup and fill _workers.
        for k, worker in list(self._workers.items()):
            if k < epoch:
                worker.exit()
                del self._workers[k]
        for ep_ in range(epoch, epoch + self.preload_next_n_sub_epochs + 1):
            if ep_ in self._workers:
                continue
            full_epoch_0idx_ = (ep_ - 1) // self.partition_epoch
            files_order: List[List[FileTree]] = self._files_order_cache[full_epoch_0idx_]
            files_for_subep = files_order[(ep_ - 1) % self.partition_epoch]
            print(f"{self}: using files for epoch {ep_}: {files_for_subep}", file=log.v4)
            dataset_dict = self._get_sub_dataset_dict(files=files_for_subep)
            worker = _WorkerProcParent(
                name=f"{self.__class__.__name__} {self.name} ep {epoch}",
                epoch=ep_,
                full_epoch_0idx=full_epoch_0idx_,
                dataset_dict=dataset_dict,
                buffer_size=self.buffer_size,
            )
            self._workers[ep_] = worker

        self._num_seqs = self._workers[epoch].get_num_seqs()
        return True

    def _get_sub_dataset_dict(self, files: List[FileTree]) -> Dict[str, Any]:
        import tree

        dataset_dict = self.get_sub_epoch_dataset(files)
        dataset_dict = extend_dataset_dict_from_parent_dataset(dataset_dict, parent_dataset=self)

        flat_sub_dset = tree.flatten_with_path(dataset_dict)

        part_epoch_cfg = next(
            ((path, v) for path, v in flat_sub_dset if path[-1] == "partition_epoch" and v != 1), None
        )
        if part_epoch_cfg is not None:
            path, subeps = part_epoch_cfg
            raise ValueError(
                f"{self}: sub dataset should not have partition_epoch, "
                f'but got "partition_epoch": {subeps} at {".".join(path)} in {dataset_dict}.'
            )

        # Heuristic check for well-definedness of seq ordering. Might need to be extended in the
        # future if there are other ways of defining a seq order than the ones below.
        if (
            not any(path[-1] == "seq_ordering" for path, _ in flat_sub_dset)
            and not any(path[-1] == "seq_order_control_dataset" for path, _ in flat_sub_dset)
            and not any(path[-1] == "map_seq_stream" for path, _ in flat_sub_dset)
        ):
            raise ValueError(
                f"{self}: there should be an explicit seq_ordering somewhere in the sub dataset "
                f"(or seq_order_control_dataset for MetaDataset or map_seq_stream for PostprocessingDataset), "
                f"but found none in {dataset_dict}."
            )

        return dataset_dict

    @staticmethod
    def _distribute_evenly_by_size(
        *, num_bins: int, file_sizes: Dict[str, int], files_order: Sequence[FileTree]
    ) -> List[List[FileTree]]:
        """
        Distributes the files from files_order into ``num_bins`` while attempting
        to make every bin as evenly sized (based on ``file_sizes``) as possible.
        """
        total_size = sum(file_sizes[_get_key_for_file_tree(f_tree)] for f_tree in files_order)
        avg_size_per_sub_epoch = total_size / num_bins
        # Now evenly distribute the files over the bins.
        # Note that many one-pass variants of algorithms to evenly distribute
        # can end up with some empty bins,
        # so we need to make sure that this is not the case.
        # E.g. consider the seqs of size [1,1,78,120] and num_bins=4.
        # That has avg size per sub epoch 50.
        # Some simple algorithms could end up with the sub epochs
        # [[1,1], [78], [120], []] or [[1,1,78], [120], [], []].
        # Or consider [5,5]+[10]*7, num_bins=5, which has avg size 16.
        # A simple algorithm could end up with [[5,5,10], [10,10], [10,10], [10,10], []].
        # See test_DistributeFilesDataset_distribute_evenly_by_size for some test cases.
        assert len(files_order) >= num_bins
        files_per_bin = [[] for _ in range(num_bins)]
        assert len(files_per_bin) == num_bins
        bin_idx = 0
        size_taken = 0
        total_size_taken = 0
        for i, f_tree in enumerate(files_order):
            size = file_sizes[_get_key_for_file_tree(f_tree)]
            num_remaining = len(files_order) - i
            total_size_taken += size

            if num_remaining <= num_bins - bin_idx - 1:
                # All remaining sub epochs must be filled.
                assert files_per_bin[bin_idx]
                bin_idx += 1
                avg_size_per_sub_epoch = (total_size - total_size_taken) / (num_bins - bin_idx)
                files_per_bin[bin_idx].append(f_tree)
                size_taken = size
                continue
            if bin_idx == num_bins - 1:
                # We are done. Just add the rest to the last sub epoch.
                files_per_bin[bin_idx].append(f_tree)
                size_taken += size
                continue
            if size_taken + size <= avg_size_per_sub_epoch:
                files_per_bin[bin_idx].append(f_tree)
                size_taken += size
                continue
            # We should increase the sub epoch index.
            # We need to decide where to add this file, to the current or the next sub epoch.
            if not files_per_bin[bin_idx] or (
                # Better to add this file to the current sub epoch?
                abs((size_taken + size) - avg_size_per_sub_epoch) <= abs(size_taken - avg_size_per_sub_epoch)
            ):
                files_per_bin[bin_idx].append(f_tree)
                size_taken = 0
            else:
                files_per_bin[bin_idx + 1].append(f_tree)
                size_taken = size
            bin_idx += 1
            avg_size_per_sub_epoch = (total_size - total_size_taken) / (num_bins - bin_idx)
        assert all(files_per_bin)
        return files_per_bin

    def _collect_single_seq(self, seq_idx: int) -> Optional[DatasetSeq]:
        seq = self._workers[self.epoch].get_data_seq(seq_idx)
        if not seq:
            return None
        seq.seq_idx = seq_idx
        return seq

    def have_seqs(self) -> bool:
        """have seqs"""
        self._lazy_init_file_list()
        return bool(self._files)

    def finish_epoch(self, *, free_resources: bool = False):
        """finish epoch"""
        super().finish_epoch(free_resources=free_resources)
        if free_resources:
            for worker in self._workers.values():
                worker.exit()
            self._workers.clear()
            self._files_order_cache.clear()
        else:
            if self.epoch in self._workers:
                worker = self._workers.pop(self.epoch)
                worker.exit()

    def get_data_keys(self) -> List[str]:
        """data keys"""
        if self._data_keys is None:
            self._lazy_init_num_outputs()
        return self._data_keys


def _get_key_for_file_tree(t: FileTree) -> str:
    """generates a deterministic key given a file tree"""
    import tree

    return ":".join(tree.flatten(t))


def _get_rank_and_size() -> Tuple[int, int]:
    """
    :return: tuple (rank, size): the global rank and size for distributed trainings
    """

    from returnn.config import get_global_config

    config = get_global_config(raise_exception=False)
    if not config:
        return 0, 1
    if config.typed_value("torch_distributed") is not None:
        import returnn.torch.distributed

        ctx = returnn.torch.distributed.get_ctx(config=config)
        return ctx.rank(), ctx.size()
    elif config.is_true("use_horovod"):
        assert config.bool("use_tensorflow", False) or config.value("backend", "").startswith("tensorflow")

        import returnn.tf.horovod

        ctx = returnn.tf.horovod.get_ctx(config=config)
        return ctx.rank(), ctx.size()
    elif config.typed_value("__debug_dummy_distributed_rank_and_size") is not None:
        return config.typed_value("__debug_dummy_distributed_rank_and_size")
    else:
        return 0, 1


class _WorkerProcParent:
    def __init__(
        self,
        *,
        name: str,
        epoch: int,
        full_epoch_0idx: int,
        dataset_dict: Dict[str, Any],
        buffer_size: int,
    ):
        # the dataset makes sure this is set
        assert "random_seed_offset" in dataset_dict

        self.epoch = epoch
        self.full_epoch_0idx = full_epoch_0idx
        self.dataset_dict = dataset_dict

        parent_conn, child_conn = _mp.Pipe()
        self.parent_conn: mpConnection = parent_conn

        # the env will be forwarded to the child process
        with override_env_var(RANDOM_SEED_OFFSET_ENV_VAR, str(dataset_dict["random_seed_offset"])):
            self.worker_proc = _mp.Process(
                name=f"{name} worker ep {epoch}",
                target=_worker_proc_loop,
                args=(epoch, buffer_size, dataset_dict, child_conn),
                daemon=True,
            )
            self.worker_proc.start()

        # Make sure the child connection is closed here.
        # It stays open in the child, until the child dies.
        # When that happens, now any consecutive read on the pipe
        # should yield an exception -- which is what we want,
        # otherwise it would just hang.
        child_conn.close()

        # This worker operates on one sub epoch.
        # The data used here is only for this sub epoch.
        # Without shuffling, the next time we would use this data is in the next full epoch.
        # Thus, use the full epoch as the epoch here.
        # Make sure this init is async - we should not wait for this in the main process when it is not used yet.
        self.parent_conn.send(("init_seq_order", {"epoch": self.full_epoch_0idx + 1}))
        self._has_num_seqs = False
        self._num_seqs: Optional[int] = None  # remains None if unknown

    def _lazy_wait_for_init_seq_order(self):
        if self._has_num_seqs:
            return
        msg, num_seqs = self.parent_conn.recv()
        assert msg == "num_seqs" and (num_seqs is None or isinstance(num_seqs, int))
        self._num_seqs = num_seqs
        self._has_num_seqs = True

    def get_num_seqs(self) -> Optional[int]:
        """num seqs for this sub epoch"""
        self._lazy_wait_for_init_seq_order()
        return self._num_seqs

    def get_data_seq(self, seq_idx: int) -> Optional[DatasetSeq]:
        """get data seq"""
        self._lazy_wait_for_init_seq_order()
        self.parent_conn.send(("get_data_seq", {"seq_idx": seq_idx}))
        msg, data = self.parent_conn.recv()
        assert msg == "data_seq"
        return data

    def exit(self, *, join: bool = True):
        """exit"""
        self._lazy_wait_for_init_seq_order()
        self.parent_conn.send(("exit", {}))
        if join:
            self.worker_proc.join()

    def __del__(self):
        # noinspection PyBroadException
        try:
            self.exit(join=False)
        except Exception:
            pass
        else:
            try_run(self.worker_proc.join)


def _worker_proc_loop(
    epoch: int,
    buffer_size: int,
    dataset_dict: Dict[str, Any],
    parent_conn: mpConnection,
):
    if sys.platform == "linux":
        with open("/proc/self/comm", "w") as f:
            f.write(f"DFD worker {epoch}")
    better_exchook.setup_all()

    assert isinstance(epoch, int) and isinstance(buffer_size, int)
    assert isinstance(dataset_dict, dict)
    assert isinstance(parent_conn, mpConnection)

    dataset = init_dataset(dataset_dict)

    got_init_seq_order = False
    cache: deque[DatasetSeq] = deque()
    next_seq_idx = 0

    # noinspection PyShadowingNames
    def _add_to_cache():
        nonlocal next_seq_idx
        if len(cache) >= buffer_size:
            return False
        if not dataset.is_less_than_num_seqs(next_seq_idx):
            return False
        dataset.load_seqs(next_seq_idx, next_seq_idx + 1)
        seq_tag = dataset.get_tag(next_seq_idx)
        features = {data_key: dataset.get_data(next_seq_idx, data_key) for data_key in dataset.get_data_keys()}
        complete_frac = dataset.get_complete_frac(next_seq_idx, allow_only_lr_suitable=True)
        res = DatasetSeq(seq_idx=next_seq_idx, seq_tag=seq_tag, features=features, complete_frac=complete_frac)
        cache.append(res)
        next_seq_idx += 1
        return True

    # noinspection PyShadowingNames
    def _get_from_cache(seq_idx: int) -> Optional[DatasetSeq]:
        if not cache:
            return None
        if seq_idx > cache[-1].seq_idx:
            return None
        for seq in cache:
            assert seq.seq_idx <= seq_idx
            if seq.seq_idx == seq_idx:
                return seq
        assert False

    # noinspection PyShadowingNames
    def _get(seq_idx: int) -> Optional[DatasetSeq]:
        if cache and seq_idx < cache[0].seq_idx:
            raise Exception(
                f"requested seq idx {seq_idx} is smaller than cache start {cache[0].seq_idx}, cannot go backwards"
            )
        res = _get_from_cache(seq_idx)
        if res:
            return res
        if not dataset.is_less_than_num_seqs(seq_idx):
            return None
        assert next_seq_idx <= seq_idx
        while True:
            if not _add_to_cache():
                raise Exception(
                    f"buffer too small, requested seq idx {seq_idx},"
                    f" cache starts at {cache[0].seq_idx if cache else None}"
                )
            assert cache[-1].seq_idx <= seq_idx
            if cache[-1].seq_idx == seq_idx:
                return cache[-1]

    try:
        while True:
            if got_init_seq_order:
                while not parent_conn.poll():
                    if not _add_to_cache():
                        break
            msg, kwargs = parent_conn.recv()
            if msg == "exit":
                break
            elif msg == "get_data_seq":
                seq_idx = kwargs["seq_idx"]
                while cache and cache[0].seq_idx < seq_idx:
                    cache.popleft()
                res = _get(seq_idx)
                parent_conn.send(("data_seq", res))
            elif msg == "init_seq_order":
                # We are responsible to get the seq order and distrib it to all the other workers.
                dataset.init_seq_order(**kwargs)
                try:
                    num_seqs = dataset.num_seqs
                except NotImplementedError:
                    num_seqs = None
                parent_conn.send(("num_seqs", num_seqs))
                got_init_seq_order = True
                next_seq_idx = 0
                cache.clear()
            else:
                raise Exception(f"unknown msg {msg!r}")
    except KeyboardInterrupt:  # when parent dies
        pass
    except EOFError:  # when parent dies
        pass



"""
Utilities to record and playback calls to `random`.
This is useful to be able to compare different backends even when they use random numbers.

It assumes that the order of `random` calls is deterministic and consistent between backends.

The logic becomes more complex when control flow is involved like conditionals or loops.
We are also storing the control flow context in the journal.
"""

from __future__ import annotations
from typing import Optional, Union, Sequence, List, Tuple
import numpy
from dataclasses import dataclass
from returnn.tensor import Tensor, Dim, ControlFlowContext
import returnn.frontend as rf


class RandomJournal:
    """random journal. see module docstring"""

    def __init__(self):
        self._entries: List[RandomJournalEntry] = []
        self._cur_entry_idx = 0
        self._graph_reader_nodes: List[Tuple[Tensor, rf.RunCtx]] = []

    def append(
        self,
        *,
        distribution: str,
        mean: Optional[Union[int, float, Tensor]] = None,
        stddev: Optional[Union[int, float, Tensor]] = None,
        bound: Optional[Union[int, float, Tensor]] = None,
        minval: Optional[Union[int, float, Tensor]] = None,
        maxval: Optional[Union[int, float, Tensor]] = None,
        seed: Optional[Union[int, Sequence[int], numpy.ndarray]] = None,
        static: Optional[bool] = None,
        out: Optional[Tensor[numpy.ndarray]],
    ):
        """append"""
        self._entries.append(
            RandomJournalEntry(
                out=out,
                control_flow_ctx=rf.get_current_control_flow_ctx(),
                run_ctx=rf.get_run_ctx(),
                distribution=distribution,
                mean=mean,
                stddev=stddev,
                bound=bound,
                minval=minval,
                maxval=maxval,
                seed=seed,
                static=static,
            )
        )

    def get_next(self, *, new_out_template: Optional[Tensor] = None) -> RandomJournalEntry:
        """read next"""
        assert self._cur_entry_idx < len(self._entries)
        entry = self._entries[self._cur_entry_idx]
        if new_out_template is not None:
            assert new_out_template.dtype == entry.out.dtype, (
                f"random journal entry dtype mismatch,"
                f" expected {new_out_template}, got {entry.out} at index {self._cur_entry_idx}"
            )
            assert len(new_out_template.dims) == len(entry.out.dims), (
                f"random journal entry dims mismatch,"
                f" expected {new_out_template}, got {entry.out} at index {self._cur_entry_idx}"
            )
            for new_dim, old_dim in zip(new_out_template.dims, entry.out.dims):
                new_dim: Dim
                old_dim: Dim
                assert new_dim.dimension == old_dim.dimension, (
                    f"random journal entry dim mismatch,"
                    f" expected {new_out_template}, got {entry.out} at index {self._cur_entry_idx}"
                )
        self._cur_entry_idx += 1
        return entry

    def reached_end(self) -> bool:
        """reached end"""
        return self._cur_entry_idx >= len(self._entries)

    def add_graph_reader_node(self, out):
        """
        In graph mode, if reading (get_next), at cgraph construction time,
        register that we are reading from the journal
        now in the current context.
        This is used in :func:`get_recent_graph_reader_node_in_accessible_ctx`.
        """
        self._graph_reader_nodes.append((out, rf.get_run_ctx()))

    def get_graph_reader_idx(self) -> int:
        """current index"""
        return len(self._graph_reader_nodes)

    def get_recent_graph_reader_node_in_accessible_ctx(self) -> Optional[Tensor]:
        """
        From the graph reader nodes, return the most recent one which is in an accessible context.
        Accessible context means either the same, or a parent context.
        """
        cur_control_flow_ctx = rf.get_current_control_flow_ctx()
        cur_run_ctx = rf.get_run_ctx()
        for prev_out, prev_run_ctx in reversed(self._graph_reader_nodes):
            if prev_run_ctx != cur_run_ctx:
                return None
            if ControlFlowContext.is_parent_or_same(prev_out.control_flow_ctx, cur_control_flow_ctx):
                return prev_out
            consumers = rf.walk_tensor_consumers(
                prev_out,
                filter_outputs=lambda x: ControlFlowContext.is_parent_or_same(x.control_flow_ctx, cur_control_flow_ctx),
                ending_condition=lambda x: ControlFlowContext.is_parent_or_same(
                    x.control_flow_ctx, cur_control_flow_ctx
                ),
            )
            if not consumers:
                raise Exception(f"cannot handle {prev_out} in current {cur_control_flow_ctx}")
            return consumers[0]
        return None


@dataclass
class RandomJournalEntry:
    """entry"""

    out: Optional[Tensor[numpy.ndarray]]

    control_flow_ctx: Optional[ControlFlowContext]
    run_ctx: rf.RunCtx

    distribution: str
    mean: Optional[Union[int, float, Tensor]] = None
    stddev: Optional[Union[int, float, Tensor]] = None
    bound: Optional[Union[int, float, Tensor]] = None
    minval: Optional[Union[int, float, Tensor]] = None
    maxval: Optional[Union[int, float, Tensor]] = None
    seed: Optional[Union[int, Sequence[int], numpy.ndarray]] = None
    static: Optional[bool] = None



"""
Some utility functions on nested structures.
"""

from __future__ import annotations
from typing import TypeVar, Optional, Union, Sequence, Tuple, Dict
import functools
import re
import tree
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


__all__ = ["mask_nested", "gather_nested", "masked_select_nested", "masked_scatter_nested"]


T = TypeVar("T")


def mask_nested(
    s: T,
    *,
    mask: Tensor,
    mask_cpu: Optional[Tensor] = None,
    mask_value: Union[T, Tensor, float, None],
    dim_map: Optional[Dict[Dim, Dim]] = None,
    allow_dim_extension: bool = True,
) -> T:
    """
    Applies where(mask, s, mask_value) for nested structures.

    :param s:
    :param mask:
    :param mask_cpu: mask tensor for CPU. this is used e.g. for dyn dim sizes
    :param mask_value:
    :param dim_map:
    :param allow_dim_extension:
    :return: s with masked values
    """
    if dim_map is None:
        dim_map = {}
    partial_kwargs = dict(mask=mask, mask_cpu=mask_cpu, dim_map=dim_map, allow_dim_extension=allow_dim_extension)
    structures = [s]
    if type(s) is type(mask_value):  # mask_value also same nested structure?
        tree.assert_same_structure(s, mask_value)
        structures.append(mask_value)
    else:
        partial_kwargs["mask_value"] = mask_value
    tree.map_structure(functools.partial(_mask_prepare_dims, **partial_kwargs), *structures)
    return tree.map_structure(functools.partial(_mask, **partial_kwargs), *structures)


def _mask_prepare_dims(
    s: T,
    mask_value: Union[T, Tensor, float, None],
    *,
    mask: Tensor,
    mask_cpu: Optional[Tensor] = None,
    dim_map: Dict[Dim, Dim],
    allow_dim_extension: bool,
) -> T:
    if isinstance(s, Dim):
        if mask_value is None:
            return s  # not sure if always correct...
        assert isinstance(mask_value, Dim)
        if s == mask_value:
            return s
        if not allow_dim_extension:
            dim_size_dims = set()
            if s.dyn_size_ext is not None:
                dim_size_dims.update(s.dyn_size_ext.dims_set)
            if mask_value.dyn_size_ext is not None:
                dim_size_dims.update(mask_value.dyn_size_ext.dims_set)
            if not mask.dims_set.issubset(dim_size_dims):
                assert not mask.dims_set.intersection(dim_size_dims)  # not sure...
                return s
        new_dyn_size = _mask(
            s.get_size_tensor(),
            mask=mask,
            mask_cpu=mask_cpu,
            mask_value=mask_value.get_size_tensor(),
            dim_map=dim_map,
            allow_dim_extension=allow_dim_extension,
        )
        new_dim = Dim(new_dyn_size, name=_extend_dim_name(s.name))
        dim_map[s] = dim_map[mask_value] = new_dim
        return new_dim
    return s


def _mask(
    s: T,
    mask_value: Union[T, Tensor, float, None],
    *,
    mask: Tensor,
    mask_cpu: Optional[Tensor] = None,
    dim_map: Dict[Dim, Dim],
    allow_dim_extension: bool,
) -> T:
    if s is None:
        return s
    if isinstance(s, Tensor):
        if s.device == "cpu" and mask_cpu is not None:
            mask = mask_cpu
        if dim_map:
            for d in s.dims:
                if d in dim_map:
                    s = rf.replace_dim_v2(s, in_dim=d, out_dim=dim_map[d])
            if isinstance(mask_value, Tensor):
                for d in mask_value.dims:
                    if d in dim_map:
                        mask_value = rf.replace_dim_v2(mask_value, in_dim=d, out_dim=dim_map[d])
        if not allow_dim_extension and isinstance(mask_value, Tensor):
            if not s.dims_set.issuperset(mask_value.dims_set):
                return s
        if not allow_dim_extension or mask_value is None or (isinstance(mask_value, (int, float)) and mask_value == 0):
            if mask.dims_set.issubset(s.dims_set):
                return rf.where(mask, s, rf.cast(mask_value, s.dtype))
            assert not mask.dims_set.intersection(s.dims_set)  # not sure...
            return s
        assert isinstance(mask_value, (int, float, Tensor))
        return rf.where(mask, s, mask_value, allow_broadcast_all_sources=True)
    if isinstance(s, Dim):
        if mask_value is None:
            return s
        assert isinstance(mask_value, Dim)
        if s == mask_value:
            return s
        return dim_map.get(s, s)
    raise TypeError(f"_mask: unexpected {s!r} type {type(s).__name__}")


def gather_nested(s: T, *, indices: Tensor, dim_map: Optional[Dict[Dim, Dim]] = None) -> T:
    """
    This is like :func:`gather`, but for nested structures.

    :param s: nested structure
    :param indices: indices tensor. see :func:`gather`
    :param dim_map: if given, this will be updated with the new dim map
    :return: s with gathered tensors
    """
    assert indices.sparse_dim
    if dim_map is None:
        dim_map = {}
    tree.map_structure(functools.partial(_gather_prepare_dims, indices=indices, dim_map=dim_map), s)
    s = tree.map_structure(functools.partial(_gather, indices=indices, dim_map=dim_map), s)
    return s


def _gather_prepare_dims(s: T, *, indices: Tensor, dim_map: Dict[Dim, Dim]) -> T:
    if isinstance(s, Dim):
        if s.dimension is not None:  # static
            return s
        if s in dim_map:
            return dim_map[s]
        if indices.sparse_dim in s.dyn_size_ext.dims:
            new_dyn_size = _gather(s.dyn_size_ext, indices=indices, dim_map=dim_map)
            new_dim = Dim(new_dyn_size, name=_extend_dim_name(s.name))
            dim_map[s] = new_dim
            return new_dim
        return s
    # everything else ignored at this stage
    return s


def _gather(s: T, *, indices: Tensor, dim_map: Optional[Dict[Dim, Dim]] = None) -> T:
    if isinstance(s, Tensor):
        if dim_map and any(d in dim_map for d in s.dims):
            for d in s.dims:
                if d in dim_map:
                    s = rf.replace_dim_v2(s, in_dim=d, out_dim=dim_map[d])
        if indices.sparse_dim in s.dims:
            # really the default case, otherwise e.g. scalar or so, independent of beam
            s = rf.gather(s, indices=indices)
        return s
    if isinstance(s, Dim):
        if s.dimension is not None:  # static
            return s
        if dim_map and s in dim_map:
            return dim_map[s]
        assert indices.sparse_dim not in s.dyn_size_ext.dims  # not expected, should be in dim_map
        return s
    raise TypeError(f"_gather: unexpected type ({type(s)})")


def masked_select_nested(
    s: T,
    *,
    mask: Tensor,
    mask_cpu: Optional[Tensor] = None,
    dims: Sequence[Dim],
    out_dim: Optional[Dim] = None,
    dim_map: Optional[Dict[Dim, Dim]] = None,
) -> Tuple[T, Dim, Dict[Dim, Dim]]:
    """
    This is like :func:`masked_select`, but for nested structures.

    :param s: nested structure
    :param mask: mask tensor. see :func:`masked_select`
    :param mask_cpu: mask tensor for CPU. this is used e.g. for dyn dim sizes
    :param dims: dims to mask. see :func:`masked_select`
    :param out_dim: the packed out dim. see :func:`masked_select`. if not given, a new one will be created.
    :param dim_map: if given, this will be updated with the new dim map
    :return: s with masked dims, out_dim, and a newly created dim map
    """
    if out_dim is None:
        out_dim = Dim(None, name="packed_new_label")  # Flat_Batch_InBeam
    if dim_map is None:
        dim_map = {}
    tree.map_structure(
        functools.partial(
            _masked_select_prepare_dims,
            mask=mask_cpu if mask_cpu is not None else mask,
            dims=dims,
            out_dim=out_dim,
            dim_map=dim_map,
        ),
        s,
    )
    s = tree.map_structure(
        functools.partial(
            _masked_select,
            mask=mask,
            mask_cpu=mask_cpu,
            dims=dims,
            out_dim=out_dim,
            dim_map=dim_map,
        ),
        s,
    )
    return s, out_dim, dim_map


def _masked_select_prepare_dims(s, *, mask: Tensor, dims: Sequence[Dim], out_dim: Dim, dim_map: Dict[Dim, Dim]):
    if isinstance(s, Dim):
        if s.dimension is not None:  # static
            return s
        if not any(d in s.dyn_size_ext.dims for d in dims):
            return s
        if s in dim_map:
            return dim_map[s]
        new_dyn_size = _masked_select(s.dyn_size_ext, mask=mask, dims=dims, out_dim=out_dim, dim_map=dim_map)
        new_dim = Dim(new_dyn_size, name=_extend_dim_name(s.name))
        dim_map[s] = new_dim
        return new_dim
    # everything else ignored at this stage
    return s


def _masked_select(
    s: T, *, mask: Tensor, mask_cpu: Optional[Tensor] = None, dims: Sequence[Dim], out_dim: Dim, dim_map: Dict[Dim, Dim]
) -> T:
    if isinstance(s, Tensor):
        if not any(d in s.dims for d in dims):
            return s  # e.g. scalar or so, independent from dims
        if s.device == "cpu" and mask_cpu is not None:
            mask = mask_cpu
        # For the masked_select, we need that all masked dims are present, so add them if not.
        # (E.g., when we mask [batch,beam], but we only have [batch], we need to add the beam dim.)
        if any(d not in s.dims for d in dims):
            s = rf.expand_dims(s, dims=[d for d in dims if d not in s.dims])
        # The packing itself (masked_select).
        s, _ = rf.masked_select(s, mask=mask, dims=dims, out_dim=out_dim)
        # In the resulting tensor, potentially replace dims.
        # In addition to the dim replacement, we also might need to slice, as the size might be smaller.
        if any(d in dim_map for d in s.dims):
            for d in s.dims:
                if d in dim_map:
                    s, _ = rf.slice(s, axis=d, size=dim_map[d])
        return s
    if isinstance(s, Dim):
        if s.dimension is not None:  # static
            return s
        if not any(d in s.dyn_size_ext.dims for d in dims):
            return s
        assert s in dim_map
        return dim_map[s]
    raise TypeError(f"_masked_select: unexpected type ({type(s)})")


def masked_scatter_nested(
    s: T,
    backup: T,
    *,
    mask: Tensor,
    mask_cpu: Tensor,
    dims: Sequence[Dim],
    in_dim: Dim,
    masked_select_dim_map: Dict[Dim, Dim],
    masked_scatter_dim_map: Optional[Dict[Dim, Dim]] = None,
) -> T:
    """
    Reverse of :func:`masked_select_nested`.

    :param s: nested structure, where dims are packed, i.e. (in_dim,...)
    :param backup: nested structure, where we scatter into. tensors like (dims...,...)
    :param mask: mask tensor. see :func:`masked_scatter`/:func:`masked_select`
    :param mask_cpu: mask tensor for CPU. this is used e.g. for dyn dim sizes. see :func:`masked_scatter`
    :param dims: dims to mask. see :func:`masked_scatter`/:func:`masked_select`
    :param in_dim: the packed in dim. see :func:`masked_scatter`
    :param masked_select_dim_map: the dim map from :func:`masked_select_nested`.
        This describes how to map dims from s to backup.
    :param masked_scatter_dim_map: for any new dims created by this function, this will be updated
    :return: backup with s scattered in
    """
    reverse_dim_map = {v: k for k, v in masked_select_dim_map.items()}
    if masked_scatter_dim_map is None:
        masked_scatter_dim_map = {}

    tree.map_structure(
        functools.partial(
            _masked_scatter_merge_dims,
            mask=mask_cpu,
            dims=dims,
            in_dim=in_dim,
            reverse_dim_map=reverse_dim_map,
            merged_dim_map=masked_scatter_dim_map,
        ),
        s,
        backup,
    )
    s = tree.map_structure(
        functools.partial(
            _masked_scatter,
            mask=mask,
            mask_cpu=mask_cpu,
            dims=dims,
            in_dim=in_dim,
            reverse_dim_map=reverse_dim_map,
            merged_dim_map=masked_scatter_dim_map,
        ),
        s,
        backup,
    )
    return s


def _masked_scatter_merge_dims(
    s: T,
    backup: T,
    *,
    mask: Tensor,
    dims: Sequence[Dim],
    in_dim: Dim,
    reverse_dim_map: Dict[Dim, Dim],
    merged_dim_map: Dict[Dim, Dim],
) -> T:
    if isinstance(s, Dim):
        # This is slightly more complex than in the _masked_select case:
        # We need to merge the s and backup depending on the mask.
        if s in reverse_dim_map:
            s = reverse_dim_map[s]
        if s == backup:
            return s
        if s in merged_dim_map:
            return merged_dim_map[s]
        # Note: s/backup might even be static dims.
        new_size = _masked_scatter(
            s.get_size_tensor(),
            backup.get_size_tensor(),
            mask=mask,
            dims=dims,
            in_dim=in_dim,
            reverse_dim_map=reverse_dim_map,
            merged_dim_map=merged_dim_map,
        )
        assert new_size.dims_set == (
            (s.get_size_tensor().dims_set | backup.get_size_tensor().dims_set) - {in_dim}
        ) | set(dims)
        new_dim = Dim(new_size, name=backup.name)
        merged_dim_map[s] = new_dim
        merged_dim_map[backup] = new_dim
        return new_dim
    # everything else ignored at this stage
    return s


def _masked_scatter(
    s: T,
    backup: T,
    *,
    mask: Tensor,
    mask_cpu: Optional[Tensor] = None,
    dims: Sequence[Dim],
    in_dim: Dim,
    reverse_dim_map: Dict[Dim, Dim],
    merged_dim_map: Dict[Dim, Dim],
) -> T:
    if isinstance(s, Tensor):
        assert isinstance(backup, Tensor)
        if s.device == "cpu" and mask_cpu is not None:
            mask = mask_cpu
        if in_dim not in s.dims:
            s = rf.expand_dim(s, in_dim)
        # Do the reverse of _masked_select above.
        # First replace the dims back.
        if any(d in reverse_dim_map for d in s.dims):
            for d in s.dims:
                if d in reverse_dim_map:
                    s = rf.replace_dim_v2(s, in_dim=d, out_dim=reverse_dim_map[d], allow_shrink=False)
        # We also might need to replace newly merged dims, both in s and backup.
        for d in s.dims:
            if d in merged_dim_map:
                s = rf.replace_dim_v2(s, in_dim=d, out_dim=merged_dim_map[d])
        for d in backup.dims:
            if d in merged_dim_map:
                backup = rf.replace_dim_v2(backup, in_dim=d, out_dim=merged_dim_map[d])
        # The unpacking itself (reversing the masked_select, i.e. masked_scatter).
        s = rf.masked_scatter(s, backup, mask=mask, dims=dims, in_dim=in_dim)
        return s
    if isinstance(s, Dim):
        # This is slightly more complex than in the _masked_select case:
        # We need to merge the s and backup depending on the mask.
        if s in reverse_dim_map:
            s = reverse_dim_map[s]
        if s in merged_dim_map:
            return merged_dim_map[s]
        return s
    raise TypeError(f"_masked_scatter: unexpected type ({type(s)})")


def _extend_dim_name(name: str) -> str:
    # check ends with _<num>
    m = re.match(r"^(.*)_(\d+)$", name)
    if m:
        return f"{m.group(1)}_{int(m.group(2)) + 1}"
    return name + "_1"



"""
:class:`TextDictDataset`
"""

from __future__ import annotations
from typing import Optional, Union, Any, Sequence, Tuple, List, Dict
import numpy as np

from returnn.log import log
from .basic import DatasetSeq
from .cached2 import CachedDataset2
from .util.vocabulary import Vocabulary


class TextDictDataset(CachedDataset2):
    """
    This dataset can read files in the format as usually generated from RETURNN search,
    i.e. with beam like (item_format = "list_with_scores")::

        {
            seq_tag: [(score1, text1), (score2, text2), ...],
            ...
        }

    Or without beam like (item_format = "single")::

        {
            seq_tag: text,
            ...
        }

    The data keys:

        data: The single (or best) sequence (encoded via vocab).
        data_flat: for list_with_scores, all sequences concatenated (encoded via vocab), in the given order
        data_seq_lens: for list_with_scores, the sequence lengths of each seq in data_flat
        scores: for list_with_scores, the scores of each seq in data_flat
    """

    def __init__(
        self,
        *,
        filename: str,
        item_format: str = "list_with_scores",
        vocab: Union[Vocabulary, Dict[str, Any]],
        **kwargs,
    ):
        """
        :param filename: text dict file. can be gzipped.
        :param item_format: "list_with_scores" or "single"
        :param vocab: to encode the text as a label sequence. See :class:`Vocabulary.create_vocab`.
        """
        super().__init__(**kwargs)
        self.filename = filename
        self.item_format = item_format
        self.vocab = vocab if isinstance(vocab, Vocabulary) else Vocabulary.create_vocab(**vocab)
        self.num_inputs = self.vocab.num_labels
        self.num_outputs = {}
        self.labels = {}
        if item_format == "list_with_scores":
            self.num_outputs.update(
                {
                    "data": (self.vocab.num_labels, 1),
                    "data_flat": (self.vocab.num_labels, 1),
                    "data_seq_lens": (1, 1),
                    "scores": (1, 1),
                }
            )
            self.labels.update({"data_flat": self.vocab.labels})
        elif item_format == "single":
            self.num_outputs.update({"data": (self.vocab.num_labels, 1)})
            self.labels.update({"data": self.vocab.labels})
        else:
            raise ValueError(f"invalid item_format {item_format!r}")
        self._data_values: Optional[List[Union[List[Tuple[float, str]], str]]] = None  # lazily loaded
        self._seq_tags: Optional[List[str]] = None  # lazily loaded
        self._seq_order: Optional[Sequence[int]] = None  # via init_seq_order

    def _load(self):
        if self._data_values is not None:
            return

        if self.filename.endswith(".gz"):
            import gzip

            txt = gzip.GzipFile(self.filename, "rb").read()
        else:
            txt = open(self.filename, "rb").read()

        from returnn.util.literal_py_to_pickle import literal_eval

        # Note: literal_py_to_pickle.literal_eval is quite efficient.
        # However, currently, it does not support inf/nan literals,
        # so it might break for some input.
        # We might want to put a simple fallback to eval here if needed.
        # Or maybe extend literal_py_to_pickle.literal_eval to support inf/nan literals.
        try:
            data: Dict[str, Any] = literal_eval(txt)
        except Exception as exc:
            print(f"{self}: Warning: literal_py_to_pickle.literal_eval failed:", file=log.v3)
            print(f"  {type(exc).__name__}: {exc}", file=log.v3)
            print("  Fallback to eval...", file=log.v3)
            data: Dict[str, Any] = eval(txt, {"nan": float("nan"), "inf": float("inf")})
        assert data is not None
        assert isinstance(data, dict)
        assert len(data) > 0
        # Check some data.
        key, value = next(iter(data.items()))
        assert isinstance(key, str), f"{self}: expected seq tag as keys, got {key!r} ({type(key)})"  # seq tag
        if self.item_format == "single":
            assert isinstance(value, str), f"{self}: expected str ({self.item_format}), got {value!r} ({type(value)})"
        elif self.item_format == "list_with_scores":
            assert isinstance(value, list), f"{self}: expected list ({self.item_format}), got {value!r} ({type(value)})"
            assert len(value) > 0, f"{self}: expected non-empty list ({self.item_format}), got {value!r} for seq {key}"
            value0 = value[0]
            assert (
                isinstance(value0, tuple)
                and len(value0) == 2
                and isinstance(value0[0], float)
                and isinstance(value0[1], str)
            ), f"{self}: expected (score,text) tuples ({self.item_format}), got {value0!r} ({type(value0)})"
        else:
            raise ValueError(f"invalid item_format {self.item_format!r}")
        self._data_values = list(data.values())
        self._seq_tags = list(data.keys())

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """init seq order"""
        super().init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        if epoch is None and seq_list is None and seq_order is None:
            self._num_seqs = 0
            return True

        random_seed = self._get_random_seed_for_epoch(epoch=epoch)
        self.vocab.set_random_seed(random_seed)

        if self.item_format == "single":

            def _get_seq_len(i: int) -> int:
                return len(self._data_values[i])

        elif self.item_format == "list_with_scores":

            def _get_seq_len(i: int) -> int:
                values = self._data_values[i]
                return sum(len(text) for _, text in values)

        else:
            raise ValueError(f"invalid item_format {self.item_format!r}")

        if seq_order is not None:
            self._seq_order = seq_order
        elif seq_list is not None:
            raise NotImplementedError(f"{self}: seq_list not supported yet")
        else:
            self._load()
            num_seqs = len(self._data_values)
            self._seq_order = self.get_seq_order_for_epoch(epoch=epoch, num_seqs=num_seqs, get_seq_len=_get_seq_len)
        self._num_seqs = len(self._seq_order)

    def _collect_single_seq(self, seq_idx: int) -> Optional[DatasetSeq]:
        orig_seq_idx = self._seq_order[seq_idx]
        seq_tag = self._seq_tags[orig_seq_idx]
        data = self._data_values[orig_seq_idx]

        res = {}
        labels_dtype = self.get_data_dtype("data")
        if self.item_format == "single":
            res["data"] = np.array(self.vocab.get_seq(data), dtype=labels_dtype)
        elif self.item_format == "list_with_scores":
            _, best = max(data, key=lambda x: x[0])
            res["data"] = np.array(self.vocab.get_seq(best), dtype=labels_dtype)
            labels = [self.vocab.get_seq(txt) for _, txt in data]
            res["data_flat"] = np.array(sum(labels, []), dtype=labels_dtype)
            res["data_seq_lens"] = np.array([len(seq) for seq in labels], dtype=self.get_data_dtype("data_seq_lens"))
            res["scores"] = np.array([score for score, _ in data], dtype=self.get_data_dtype("scores"))
        else:
            raise ValueError(f"invalid item_format {self.item_format!r}")
        return DatasetSeq(seq_idx=seq_idx, features=res, seq_tag=seq_tag)

    def supports_sharding(self) -> bool:
        """:return: whether this dataset supports sharding"""
        return True

    def supports_seq_order_sorting(self) -> bool:
        """supports sorting"""
        return True

    def get_current_seq_order(self) -> Sequence[int]:
        """:return: seq order"""
        assert self._seq_order is not None, "init_seq_order not called"
        return self._seq_order

    def have_corpus_seq_idx(self) -> bool:
        """
        :return: whether we can use :func:`get_corpus_seq_idx`
        """
        return True

    def get_corpus_seq_idx(self, seq_idx: int) -> int:
        """
        :param seq_idx:
        """
        assert self._seq_order is not None, "init_seq_order not called"
        return self._seq_order[seq_idx]

    def get_tag(self, seq_idx: int) -> str:
        """
        :param seq_idx:
        :return: seq tag
        """
        self._load()
        return self._seq_tags[self._seq_order[seq_idx]]

    def get_all_tags(self) -> List[str]:
        """:return: all tags"""
        self._load()
        return self._seq_tags

    def get_total_num_seqs(self, *, fast: bool = False) -> int:
        """:return: total num seqs in dataset (not for (sub)epoch)"""
        self._load()
        return len(self._data_values)

    def get_data_dim(self, key: str) -> int:
        """:return: dim of data entry with `key`"""
        if key == "data" or key == "data_flat":
            return self.vocab.num_labels
        elif key == "data_seq_lens":
            return 1
        elif key == "scores":
            return 1
        else:
            raise ValueError(f"{self}: unknown data key: {key}")

    def get_data_dtype(self, key: str) -> str:
        """:return: dtype of data entry with `key`"""
        if key == "data" or key == "data_flat":
            return "int32"
        elif key == "data_seq_lens":
            return "int32"
        elif key == "scores":
            return "float32"
        else:
            raise ValueError(f"{self}: unknown data key: {key}")

    def get_data_keys(self) -> List[str]:
        """:return: available data keys"""
        return list(self.num_outputs.keys())

    def get_data_shape(self, key: str) -> List[str]:
        """
        :returns get_data(*, key).shape[1:], i.e. num-frames excluded
        """
        return []  # all are scalar or sparse

    def is_data_sparse(self, key: str) -> bool:
        """:return: whether data entry with `key` is sparse"""
        return key == "data" or key == "data_flat"



"""
Provides the :class:`LSTM` module.
"""

from __future__ import annotations

from typing import Tuple, Sequence

import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, single_step_dim


__all__ = ["LSTM", "LstmState", "ZoneoutLSTM"]


class LSTM(rf.Module):
    """
    LSTM module.
    """

    def __init__(
        self,
        in_dim: Dim,
        out_dim: Dim,
        *,
        with_bias: bool = True,
    ):
        """
        Code to initialize the LSTM module.
        """
        super().__init__()

        self.in_dim = in_dim
        self.out_dim = out_dim

        self.ff_weight = rf.Parameter((4 * self.out_dim, self.in_dim))
        self.ff_weight.initial = rf.init.Glorot()
        self.rec_weight = rf.Parameter((4 * self.out_dim, self.out_dim))
        self.rec_weight.initial = rf.init.Glorot()

        self.bias = None
        if with_bias:
            self.bias = rf.Parameter((4 * self.out_dim,))
            self.bias.initial = 0.0

    def __call__(self, source: Tensor, *, state: LstmState, spatial_dim: Dim) -> Tuple[Tensor, LstmState]:
        """
        Forward call of the LSTM.

        :param source: Tensor of size {...,in_dim} if spatial_dim is single_step_dim else {...,spatial_dim,in_dim}.
        :param state: State of the LSTM. Both h and c are of shape {...,out_dim}.
        :return: output of shape {...,out_dim} if spatial_dim is single_step_dim else {...,spatial_dim,out_dim},
            and new state of the LSTM.
        """
        if state.h is None or state.c is None:
            raise ValueError(f"{self}: state {state} needs attributes ``h`` (hidden) and ``c`` (cell).")
        if self.in_dim not in source.dims:
            raise ValueError(f"{self}: input {source} does not have in_dim {self.in_dim}")

        # noinspection PyProtectedMember
        result, (new_state_h, new_state_c) = source._raw_backend.lstm(
            source=source,
            state_c=state.c,
            state_h=state.h,
            ff_weight=self.ff_weight,
            rec_weight=self.rec_weight,
            bias=self.bias,
            spatial_dim=spatial_dim,
            in_dim=self.in_dim,
            out_dim=self.out_dim,
        )
        new_state = LstmState(h=new_state_h, c=new_state_c)
        result.feature_dim = self.out_dim
        new_state.h.feature_dim = self.out_dim
        new_state.c.feature_dim = self.out_dim

        return result, new_state

    def default_initial_state(self, *, batch_dims: Sequence[Dim]) -> LstmState:
        """initial state"""
        return LstmState(
            h=rf.zeros(list(batch_dims) + [self.out_dim], feature_dim=self.out_dim),
            c=rf.zeros(list(batch_dims) + [self.out_dim], feature_dim=self.out_dim),
        )


class LstmState(rf.State):
    """LSTM state"""

    def __init__(self, *_args, h: Tensor = None, c: Tensor = None):
        super().__init__(*_args)
        if not _args:
            self.h = h
            self.c = c


class ZoneoutLSTM(LSTM):
    """
    Zoneout LSTM module.
    """

    def __init__(
        self,
        in_dim: Dim,
        out_dim: Dim,
        *,
        with_bias: bool = True,
        zoneout_factor_cell: float = 0.0,
        zoneout_factor_output: float = 0.0,
        use_zoneout_output: bool = True,
        forget_bias: float = 0.0,
        parts_order: str = "ifco",
    ):
        """
        :param in_dim:
        :param out_dim:
        :param with_bias:
        :param zoneout_factor_cell: 0.0 is disabled. reasonable is 0.15.
        :param zoneout_factor_output: 0.0 is disabled. reasonable is 0.05.
        :param use_zoneout_output: True is like the original paper. False is like older RETURNN versions.
        :param forget_bias: 1.0 is default in RETURNN/TF ZoneoutLSTM.
            0.0 is default in :class:`LSTM`, or RETURNN NativeLSTM, PyTorch LSTM, etc.
        :param parts_order:
            i: input gate.
            f: forget gate.
            o: output gate.
            c|g|j: input.
            icfo: like RETURNN/TF ZoneoutLSTM.
            ifco: PyTorch (cuDNN) weights, standard for :class:`LSTM`.
            cifo: RETURNN NativeLstm2 weights.
        """
        super().__init__(
            in_dim,
            out_dim,
            with_bias=with_bias,
        )
        self.zoneout_factor_cell = zoneout_factor_cell
        self.zoneout_factor_output = zoneout_factor_output
        self.use_zoneout_output = use_zoneout_output
        self.forget_bias = forget_bias
        self.parts_order = parts_order.replace("c", "j").replace("g", "j")
        self.dropout_broadcast = rf.dropout_broadcast_default()
        assert len(self.parts_order) == 4 and set(self.parts_order) == set("ijfo")

    def _inner_step(self, x: Tensor, *, state: LstmState) -> Tuple[Tensor, LstmState]:
        prev_c = state.c
        prev_h = state.h

        # Apply vanilla LSTM
        rec = rf.dot(prev_h, self.rec_weight, reduce=self.out_dim)
        x = x + rec
        parts = rf.split(x, axis=4 * self.out_dim, out_dims=[self.out_dim] * 4)
        parts = {k: v for k, v in zip(self.parts_order, parts)}
        i, j, f, o = parts["i"], parts["j"], parts["f"], parts["o"]

        new_c = rf.sigmoid(f + self.forget_bias) * prev_c + rf.sigmoid(i) * rf.tanh(j)
        new_h = rf.sigmoid(o) * rf.tanh(new_c)
        output = new_h

        # Now the ZoneoutLSTM part, which is optional (zoneout_factor_cell > 0 or zoneout_factor_output > 0).
        c = _zoneout(
            prev=prev_c,
            cur=new_c,
            factor=self.zoneout_factor_cell,
            out_dim=self.out_dim,
            dropout_broadcast=self.dropout_broadcast,
        )
        h = _zoneout(
            prev=prev_h,
            cur=new_h,
            factor=self.zoneout_factor_output,
            out_dim=self.out_dim,
            dropout_broadcast=self.dropout_broadcast,
        )
        new_state = LstmState(c=c, h=h)

        if self.use_zoneout_output:  # really the default, sane and original behavior
            output = h

        output.feature_dim = self.out_dim
        new_state.h.feature_dim = self.out_dim
        new_state.c.feature_dim = self.out_dim
        return output, new_state

    def __call__(self, source: Tensor, *, state: LstmState, spatial_dim: Dim) -> Tuple[Tensor, LstmState]:
        """
        Forward call of the LSTM.

        :param source: Tensor of size {...,in_dim} if spatial_dim is single_step_dim else {...,spatial_dim,in_dim}.
        :param state: State of the LSTM. Both h and c are of shape {...,out_dim}.
        :return: output of shape {...,out_dim} if spatial_dim is single_step_dim else {...,spatial_dim,out_dim},
            and new state of the LSTM.
        """
        if state.h is None or state.c is None:
            raise ValueError(f"{self}: state {state} needs attributes ``h`` (hidden) and ``c`` (cell).")
        if self.in_dim not in source.dims:
            raise ValueError(f"{self}: input {source} does not have in_dim {self.in_dim}")

        x = rf.dot(source, self.ff_weight, reduce=self.in_dim)
        if self.bias is not None:
            x = x + self.bias

        if spatial_dim == single_step_dim:
            return self._inner_step(x, state=state)

        batch_dims = source.remaining_dims((spatial_dim, self.in_dim))
        output, new_state, _ = rf.scan(
            spatial_dim=spatial_dim,
            initial=state,
            xs=x,
            ys=Tensor("lstm-out", dims=batch_dims + [self.out_dim], dtype=source.dtype, feature_dim=self.out_dim),
            body=lambda x_, s: self._inner_step(x_, state=s),
        )
        return output, new_state


def _zoneout(*, prev: Tensor, cur: Tensor, factor: float, out_dim: Dim, dropout_broadcast: bool) -> Tensor:
    if factor == 0.0:
        return cur
    return rf.cond(
        rf.get_run_ctx().is_train_flag_enabled(func=ZoneoutLSTM.__call__),
        lambda: (1 - factor) * rf.dropout(cur - prev, factor, axis=dropout_broadcast and out_dim) + prev,
        lambda: (1 - factor) * cur + factor * prev,
    )
